/**	====================================================================  *
 ||	Module:		RoomWrek.CPP
 ||	Subject:	RoomWreck v0.96
 ||	Author:		Golden Spud
 ||	--------------------------------------------------------------------
 ||	Notes:
 || Manages roomreq.??? files using following steps:
 ||		1.	reads room.dat (GremCit64/Cit+ format) and netid.cit
 || 	2.	compares them, gets rid of rooms that exist (keeps
 ||			a list of deleted rooms, in other words)
 || 	3.	reads roomreq file
 || 	4.	compares deleted rooms and roomreq file, getting rid of
 ||			all deleted rooms from roomreq
 || 	5.	writes out new roomreq that does not have deleted rooms
 || Futures:
 ||		[most, if not all, Zookeeper's ideas:]
 ||		-	write out deleted rooms:
 ||				roomwrek deleted room.dat netid.cit output.fil
 ||		-	normal operation:
 ||				roomwrek timesave room.dat netid.cit roomreq.xxx
 ||		-	not in roomreq:
 ||				roomwrek norequest room.dat netid.cit roomreq.xxx output.fil
 ||		-	write missing rooms to netid.cit
 ||				roomwrek fixnetid room.dat netid.cit
 ||     -   figure out what's wrong with roomwreck and/or zookie's setup
 ||	--------------------------------------------------------------------
 ||	History:	(insert new entries at top of list)
 ||	YYmmmDD/nam	description
 ||
 || 96mar27/ms  released version 0.96
 || 96mar26/ms	moved the entire upper-level process to timeSave()
 || 96mar26/ms	finished moving necessary junk to separate functions
 || 96mar26/ms	moved all of the code that differed between the set and
 || 			 the vector builds into macros to unclutter code and
 || 			 facilitate maintenance
 || 96mar26/ms	made exception handling optional during compile time
 || 			 because I was having crash problems with it (even though no
 || 			 exceptions were being thrown).
 || 96mar26/ms  somehow made the USE_SET option actually work. FASTER!
 || 96feb??/ms	started using C++ exception handling to handle errors
 || 			 in the separated functions
 || 96feb??/ms	started moving stuff into separate functions in order to
 || 			 prepare for command-line options for extra administrative
 || 			 features
 || 96feb27/ms 	added a counter when reading files to show what number
 || 			 it's reading
 || 96feb27/ms 	added USE_SET compile option which attempts to use sets
 || 			 instead of vectors so that sorting doesn't have to be done
 || 96feb27/ms 	changed netid.cit reading method. It used to read netid.cit
 || 			 slower than it would sort it. Now it reads it much faster.
 || 96feb27/ms 	changed room.dat reading to work with any version of
 || 			 GremCit/Cit+ [Brent Bottles]
 ||  
 || 96feb22/ms 	released version 0.9
 || 96feb22/ms 	started development
 **	====================================================================  */

// ======================================================= Includes ===
#include <iostream.h>
#include <fstream.h>
#include <string.h>
#include <string.hpp>	   // WATCOM C++ dependent
#include <vector.h>
#include <set.h>
#include <algo.h>
#include <bool.h>

#include "genlib.h"
#include "types.h"

// ======================================================== Defines ===
//#define USE_SET			// If on, uses sets instead of vectors
//#define EXCEPTIONS		// 16-bit compile seems to have problems w/ these

// ----------------------
// The following macros are defined to clean up the rest of the code.
// main() was becoming a mess of preprocessor directives and the code
// was becoming very hard to read.

// This one exits the program with an error if an exception was thrown,
// either using the C++ exception handling stuff or the home-grown isError
// method.

#ifdef EXCEPTIONS
	#define TRY(x)									\
		{											\
		try											\
			{										\
			x;										\
			}										\
		catch (String errorStr)						\
			{										\
			cerr << "Error: " << errorStr << "\n";	\
			exit(1);								\
			}										\
		}
#else
	#define TRY(x)									\
		{											\
		isError = false;							\
		x;											\
													\
		if (isError)								\
			{										\
			cerr << "Error: " << errorStr << "\n";	\
			exit(1);								\
			}										\
		}
#endif

// This one throws an exception from a function returning void, either using
// the C++ exception handling stuff or the home-grown isError method.

#ifdef EXCEPTIONS
	#define THROW(x)	\
		{				\
		errorStr = (x);	\
		throw errorStr;	\
		}
#else
	#define THROW(x)			\
		{						\
		String errorStr = (x);	\
		isError = true;			\
		return;					\
		}
#endif

// ---------------------
// The following is just shorthand because this type of comparison
// seems to be done a lot. In z, it places the strings that are in
// x but not in y.

// this keeps it to one macro which is easier to maintain
#ifdef USE_SET
	#define INSERTER set_inserter
#else
	#define INSERTER back_inserter
#endif

// the macro (ta-da!)
#define DIFF(x,y,z)							\
	{										\
	set_difference((x).begin(), (x).end(),	\
				   (y).begin(), (y).end(),	\
				   INSERTER((z)),			\
				   strless_i());			\
	}
	
// --------------------
// More shorthand -- this time, it's for pushing a string into the
// container. Since sets don't have a push_back function due to their
// design, it uses insert with one argument. Since vectors have insert,
// but it takes two arguments, I can't just use insert. Thus, my code
// became riddled with a bunch of #ifdef USE_SET structures. This macro
// prevents that. It pushes y into x.

#ifdef USE_SET
	#define PUSH(x,y)		\
		{					\
		(x).insert((y));	\
		}
#else
	#define PUSH(x,y)		\
		{					\
		(x).push_back((y));	\
		}
#endif

// -----------------
// Yet more shorthand -- this sorts a container. Of course, since
// sets don't need to be sorted, we don't sort them! Since the output
// before the sort always shows the name of the container being sorted,
// we take advantage of the ANSI-C stringizing construct

#ifndef USE_SET
	#define SORT(x)									\
		{											\
		cout << "Sorting " << #x << "...\n";		\
		sort((x).begin(), (x).end(), strless_i());	\
		}
#else
	#define SORT(x)
#endif

// ====================================================== Constants ===
const char* banner =
	"RoomWreck v0.96 -- roomreq.??? file manager extraordinaire!\n"
	"by Golden Spud @ Golden Spud's Greasehouse.\n";

const word BM1_INUSE =			0x0001;
const word BM1_PUBLIC =			0x0002;
const word BM1_DIRECTORY =		0x0004;
const word BM1_PERMANENT =		0x0008;
const word BM1_GROUPONLY =		0x0010;
const word BM1_READONLY =		0x0020;
const word BM1_DOWNLOADONLY =	0x0040;
const word BM1_SHARED =			0x0080;
const word BM1_MODERATED =		0x0100;
const word BM1_APPLICATION =	0x0200;
const word BM1_BIO =			0x0400;
const word BM1_UPLOADONLY =		0x0800;
const word BM1_PRIVGRP =		0x1000;
const word BM1_ANONYMOUS =		0x2000;
const word BM1_ASKSUBJECT =		0x4000;
const word BM1_GRPMODERATED =	0x8000;

const word BM2_ARCHIVEMSGS =	0x0001;
const word BM2_BOOLGROUP =		0x0002;

const word BM3_AUTOAPP =		0x0001;
const word BM3_NOEXCLUDE =		0x0002;

// =============================================== Function Objects ===

class strless_i 
	{
	public:
		bool operator()(const string& a, const string& b) const 
			{
			return a.lower() < b.lower();
			}
	};

// ========================================================== Types ===
typedef String string;

// tRoomDat contains only the fields that are necessary for RoomWreck
// to function. The rest of the fields are left out with a preprocessor
// structure in case they are needed in the future for something.
struct tRoomDat 
	{
	word Bitmapped_1;
	word Bitmapped_2;
	char Name[31];
	char DirName[64];
	char RoomTell[13];
	char ApplicName[13];
	byte Unused_2;
	word GroupNumber;
	dword Unused_3;
	word PrivGrpNum;
	word Bitmapped_3;
	char Infoline[80];
	char NetID[31];
#ifdef NEED_FULL_TROOMDAT
	char ArchiveFile[13];
	char Dictionary[13];
	char Moderator[31];
	char ExclNetPartner[31];
	byte Unused_5;
	byte BoolGroup[52];
#endif // NEED_FULL_TROOMDAT
	};

#ifndef USE_SET
	typedef vector<string> stringcontainer;
#else
	typedef set<string, strless_i> stringcontainer;

	template <class Container>
	class set_insert_iterator : public output_iterator 
		{
		protected:
			Container& container;
		public:
			set_insert_iterator(Container& x) : container(x) 
				{
				}
				
			set_insert_iterator<Container>&
			operator=(const Container::value_type& value) 
				{
				container.insert(value);
				return *this;
				}
				
			set_insert_iterator<Container>& operator*()
				{
				return *this;
				}
				
			set_insert_iterator<Container>& operator++()
				{
				return *this;
				}
				
			set_insert_iterator<Container>& operator++(int)
				{
				return *this;
				}
		};
#endif // USE_SET

// ===================================================== Prototypes ===
#ifdef USE_SET
	template <class Container>
	set_insert_iterator<Container> set_inserter(Container& c) 
		{
		return set_insert_iterator<Container>(c);
		}
#endif

void getRoomDat(string fn, stringcontainer& vRoomDat);
void getNetidCit(string fn, stringcontainer& vNetidCit);
void getRoomreq(string fn, stringcontainer& vRoomreq);
void timeSave(string fnRoomDat, string fnNetidCit, string fnRoomreq);

// =============================================== Global Variables ===

// ================================================ Local Variables ===
#ifndef EXCEPTIONS
	static String errorStr;
	static bool isError;
#endif

// ================================================= Implementation ===
int main(int argc, char** argv)
	{
	cout << banner;
	
	if (argc != 4) 
		{
		cerr << "Command-line format is:\n"
			 << "  " << argv[0] << " room.dat netid.cit roomreq.abc\n"
			 << "room.dat    is the path and filename of your room.dat,\n"
			 << "netid.cit   is the path and filename of your netid.cit,\n"
			 << "roomreq.abc is the path and filename of the roomreq "
			 << "to process.\n";
		return 1;
		}

	string fnRoomDat = argv[1];
	string fnNetidCit = argv[2];
	string fnRoomreq = argv[3];
	
	TRY(timeSave(fnRoomDat, fnNetidCit, fnRoomreq));

	// ---- Done

	cout << "Done with operation!\n";

	return 0;
	}



//------------
void getRoomDat(string fn, stringcontainer& vRoomDat) 
	{
	cout << "Reading in room.dat...";

	ifstream fRoomDat(fn, ios::in | ios::binary);
	if (!fRoomDat) 
		{
		string err = fn + " cannot be opened.";
		THROW(err);
		}

	dword reclen;
	fRoomDat.read((char*)&reclen, sizeof (reclen));
	if (reclen < sizeof (tRoomDat)) 
		{
		string err = "room.dat is in an invalid format.";
		THROW(err);
		}

	// allocate buffer needed to store rest of crap in room.dat entries
	char* buffer = new char[reclen - sizeof (tRoomDat)];
	if (!buffer) 
		{
		string err = "could not allocate buffer.";
		THROW(err);
		}
	
	tRoomDat RoomInfo;
	for (fRoomDat.read((char*)&RoomInfo, sizeof (RoomInfo)),
			 fRoomDat.read(buffer, size_t(reclen) - sizeof (RoomInfo));
		 fRoomDat;
		 fRoomDat.read((char*)&RoomInfo, sizeof (RoomInfo)),
			 fRoomDat.read(buffer, size_t(reclen) - sizeof (RoomInfo)))
		{
		if (RoomInfo.Bitmapped_1 & BM1_INUSE)
			{
			if (RoomInfo.Bitmapped_1 & BM1_SHARED)
				{
				PUSH(vRoomDat, RoomInfo.NetID);
				}
			}
				
		#ifndef NDEBUG
			cout << "\n" << RoomInfo.NetID;
		#else
			if (vRoomDat.size() % 25 == 0)
				{
				cout << "\rReading in room.dat... " << vRoomDat.size();
				}
		#endif
		}
	cout << "\rReading in room.dat... " << vRoomDat.size() << " net rooms.\n";
	fRoomDat.close();

	SORT(vRoomDat);
}



// ------------
void getNetidCit(string fn, stringcontainer& vNetidCit)
	{
	cout << "Reading in netid.cit...";

	ifstream fNetidCit(fn);
	if (!fNetidCit) 
		{
		string err = fn + " cannot be opened.";
		THROW(err);
		}

	static char scratch[128];
	for (fNetidCit.getline(scratch, 127); fNetidCit;
		 	fNetidCit.getline(scratch, 127))
		{
		if (scratch[strlen(scratch) - 1] == '\n')
			{
			scratch[strlen(scratch) - 1] = 0;
			}

		string sNetid(30, string::reserve);
			
		if (scratch[0] != '#')
			{
			continue;
			}

		// find first quote mark
		char *pScratch;
		for (pScratch = scratch; *pScratch != '\"'; ++pScratch)
			{
			}

		// skip first quote mark, add to string 'til next quote mark
		for (++pScratch; *pScratch != '\"'; ++pScratch)
			{
			sNetid += *pScratch;
			}

		// discard rest, push the net id into the vector
		PUSH(vNetidCit, sNetid);

		#ifndef NDEBUG
			cout << "\n" << sNetid;
		#else
			if (vNetidCit.size() % 25 == 0)
				{
				cout << "\rReading in netid.cit... " << vNetidCit.size();
				}
		#endif
		}
	cout << "\rReading in netid.cit... " << vNetidCit.size() << " net ID's\n";
	fNetidCit.close();

	SORT(vNetidCit);
	}



// -----------
void getRoomreq(string fn, stringcontainer& vRoomreq)
	{
	cout << "Reading in room request " << fn << "...";

	ifstream fRoomreq(fn);
	if (!fRoomreq) 
		{
		string err = fn + " cannot be opened.";
		THROW(err);
		}

	static char scratch[128];
	for (fRoomreq.getline(scratch, 127); fRoomreq;
			 fRoomreq.getline(scratch, 127))
		{
		// this first strlen compare prevents scratch[-1] below
		if (strlen(scratch) == 0)
			{
			continue;
			}
		
		if (scratch[strlen(scratch) - 1] == '\n')
			{
			scratch[strlen(scratch) - 1] = 0;
			}

		if (strlen(scratch) == 0)
			{
			continue;
			}

		PUSH(vRoomreq, scratch);

		#ifndef NDEBUG
			cout << "\n" << scratch;
		#else
			if (vRoomreq.size() % 25 == 0)
				{
				cout << "\rReading in room request " << fn << "... "
					 << vRoomreq.size();
				}
		#endif
		}
	cout << "\rReading in room request " << fn << "... " << vRoomreq.size()
		 << " room requests.\n";
	fRoomreq.close();

	SORT(vRoomreq);
	}



// -----------------
void timeSave(string fnRoomDat, string fnNetidCit, string fnRoomreq)
	{
	// ---- 1a. read room.dat

	stringcontainer vRoomDat;
	TRY(getRoomDat(fnRoomDat, vRoomDat));

	// ---- 1b. read netid.cit
	
	stringcontainer vNetidCit;
	TRY(getNetidCit(fnNetidCit, vNetidCit));
	
	// ---- 2. keep only deleted rooms

	cout << "Comparing vRoomDat and vNetidCit...";

	stringcontainer vDeleted;
	DIFF(vNetidCit, vRoomDat, vDeleted);

	cout << " " << vDeleted.size() << " deleted rooms.\n";

	#ifndef NDEBUG
		for (stringcontainer::iterator vDi = vDeleted.begin();
				 vDi != vDeleted.end(); ++vDi)
			{
			cout << "\n" << *vDi;
			}
	#endif

	// ---- 3. read in roomreq file
	
	stringcontainer vRoomreq;
	TRY(getRoomreq(fnRoomreq, vRoomreq));
	
	// ---- 4. compare deleted rooms and roomreq file

	cout << "Comparing vRoomreq and vDeleted...";

	stringcontainer vNotDeleted;
	DIFF(vRoomreq, vDeleted, vNotDeleted);
	
	cout << " " << vNotDeleted.size() << " rooms left.\n";

	#ifndef NDEBUG
		for (stringcontainer::iterator vNDi = vNotDeleted.begin();
				vNDi != vNotDeleted.end(); ++vNDi)
			{
			cout << "\n" << *vNDi;
			}
	#endif

	// ---- 5. Write out new roomreq file

	cout << "Writing new room request " << fnRoomreq << "...";

	ofstream oRoomreq(fnRoomreq, ios::out | ios::truncate);
	if (!oRoomreq) 
		{
		string err = fnRoomreq + " cannot be opened.";
		THROW(err);
		}

	for (stringcontainer::iterator i = vNotDeleted.begin();
			i != vNotDeleted.end(); ++i)
		{
		oRoomreq << "\n" << *i;
		}
	cout << "\n";
	oRoomreq.close();

	}

// ==================================================== End of File ===
